(function() {
  var Entry, Logger, defaultLogger, defaults, esc, fs, m, middleware, outputters, util;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice;
  util = require('util');
  fs = require('fs');
  defaults = {};
  defaults.levels = {
    panic: 0,
    error: 1,
    warn: 2,
    info: 3,
    debug: 4,
    trace: 5
  };
  defaults.level = 'info';
  defaults.fileFormat = "%timestamp% [%level%] %message%\n";
  defaults.consoleFormat = "[%(default)level%] %(default)message%%:n:inspect%";
  defaults.colors = {
    black: 30,
    red: 31,
    green: 32,
    yellow: 33,
    blue: 34,
    magenta: 35,
    cyan: 36,
    white: 37,
    bold: 1,
    underline: 4,
    reversed: 7
  };
  defaults.colorMap = {
    info: 'white',
    debug: 'cyan',
    warn: 'yellow',
    error: 'red'
  };
  middleware = m = {};
  middleware.color = function(map, colors, defaultColor) {
    var fn;
    if (map == null) {
      map = defaults.colorMap;
    }
    if (colors == null) {
      colors = defaults.colors;
    }
    if (defaultColor == null) {
      defaultColor = 'white';
    }
    fn = function(entry) {
      return map[entry.level] || defaultColor;
    };
    fn.colors = colors;
    return fn;
  };
  middleware.inspector = function(type) {
    var types;
    if (type == null) {
      type = 'inspect';
    }
    types = {
      inspect: function(param) {
        return util.inspect(param);
      },
      json: function(param) {
        return JSON.stringify(param, null, '  ');
      }
    };
    return function(entry) {
      return entry.get('params').map(function(param) {
        return types[type](param);
      }).join('\n');
    };
  };
  middleware.datetime = function() {
    return function(entry) {
      var d;
      d = new Date();
      return ("" + (d.getFullYear()) + (d.getMonth() + 1) + (d.getDate()) + " ") + ("" + (d.getHours()) + ":" + (d.getMinutes()) + ":" + (d.getSeconds()));
    };
  };
  outputters = {
    stdout: function(opts) {
      if (opts == null) {
        opts = {};
      }
      opts.format || (opts.format = defaults.consoleFormat);
      return function(entry) {
        if (!opts.level || entry.logger.shouldLog(entry.level, opts.level)) {
          return console.log(entry.format(opts.format));
        }
      };
    },
    stderr: function(opts) {
      if (opts == null) {
        opts = {};
      }
      opts.format || (opts.format = defaults.consoleFormat);
      return function(entry) {
        if (!opts.level || entry.logger.shouldLog(entry.level, opts.level)) {
          return console.warn(entry.format(opts.format));
        }
      };
    },
    file: function(opts) {
      var stream;
      if (opts == null) {
        opts = {};
      }
      opts.format || (opts.format = defaults.fileFormat);
      if (opts.filename) {
        stream = fs.createWriteStream(opts.filename);
      }
      stream || (stream = opts.stream);
      if (!stream) {
        throw new Error('Cannot log to file without filename or stream.');
      }
      return function(entry) {
        if (!opts.level || entry.logger.shouldLog(entry.level, opts.level)) {
          return stream.write(entry.format(opts.format));
        }
      };
    }
  };
  defaults.attrs = {
    timestamp: function() {
      return new Date().toISOString();
    },
    datetime: function() {
      return m.datetime();
    },
    inspect: m.inspector(),
    color: m.color()
  };
  defaults.outputs = [outputters.stdout()];
  esc = function(str) {
    return "\x1B[" + str + 'm';
  };
  Entry = (function() {
    function Entry(options) {
      var k;
      for (k in options) {
        this[k] = options[k];
      }
    }
    Entry.prototype.get = function(param) {
      if (this.logger.attrs[param]) {
        return this.logger.attrs[param](this);
      } else {
        return this[param];
      }
    };
    Entry.prototype.format = function(formatstring) {
      return formatstring.replace(/%(\(([\w,]+)\))?(:([snt]+):)?(\w+)%/g, __bind(function(str, p1, color, p3, space, param, offset, s) {
        var colors, r;
        r = this.get(param);
        if (color && this.get('color')) {
          colors = this.logger.attrs['color'].colors;
          r = esc(colors[color === 'default' ? this.get('color') : color]) + r + esc(0);
        }
        if (space && r) {
          r = space.replace(/n/g, '\n').replace(/s/g, ' ').replace(/t/g, '\t') + r;
        }
        return r;
      }, this));
    };
    return Entry;
  })();
  Logger = (function() {
    function Logger(options, attrs, _outputs) {
      var key, val, _ref;
      if (options == null) {
        options = {};
      }
      this.attrs = attrs != null ? attrs : defaults.attrs;
      this._outputs = _outputs != null ? _outputs : defaults.outputs;
      _ref = {
        levels: defaults.levels,
        level: defaults.level,
        modules: {}
      };
      for (key in _ref) {
        val = _ref[key];
        if (!options[key]) {
          options[key] = val;
        }
      }
      this._opts = {};
      this.set(options);
      this.defaults = defaults;
      this.middleware = this.m = middleware;
      this.outputters = this.out = outputters;
    }
    Logger.prototype.set = function(options) {
      if (options.levels) {
        Object.keys(this._opts.levels || {}).forEach(__bind(function(level) {
          delete this[level];
          return delete this[level[0].toUpperCase()];
        }, this));
        Object.keys(options.levels).forEach(__bind(function(level) {
          var first;
          this[level] = __bind(function() {
            var message, params;
            message = arguments[0], params = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
            if (this.shouldLog(level, this._opts.level)) {
              return this._log({
                logger: this,
                level: level,
                message: message,
                params: params,
                module: ''
              });
            }
          }, this);
          first = level[0].toUpperCase();
          if (!this[first]) {
            return this[first] = this[level];
          }
        }, this));
        this._opts.levels = options.levels;
      }
      if (options.modules) {
        this._opts.modules = options.modules;
      }
      if (options.level) {
        this._opts.level = options.level;
      }
      return this;
    };
    Logger.prototype.shouldLog = function(thisLevel, loggableLevel) {
      return this._opts.levels[thisLevel] <= this._opts.levels[loggableLevel];
    };
    Logger.prototype.log = function() {
      var level, message, params;
      level = arguments[0], message = arguments[1], params = 3 <= arguments.length ? __slice.call(arguments, 2) : [];
      if (this.shouldLog(level, this._opts.level)) {
        return this._log({
          logger: this,
          level: level,
          message: message,
          params: params,
          module: ''
        });
      }
    };
    Logger.prototype._log = function(options) {
      var entry;
      entry = new Entry(options);
      return this._outputs.forEach(function(out) {
        return out(entry);
      });
    };
    Logger.prototype.to = function() {
      var outputters;
      outputters = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      this._outputs = outputters;
      return this;
    };
    Logger.prototype.module = function(module) {
      var moduleLogger;
      moduleLogger = {};
      Object.keys(this._opts.levels).forEach(__bind(function(level) {
        var first;
        moduleLogger[level] = __bind(function() {
          var message, modulelevel, params;
          message = arguments[0], params = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
          modulelevel = this._opts.modules[module];
          if (!modulelevel && module.split('::').length > 1) {
            modulelevel = this._opts.modules[module.split('::')[0]];
          }
          if (this.shouldLog(level, modulelevel || this._opts.level)) {
            return this._log({
              logger: this,
              level: level,
              message: message,
              params: params,
              module: module
            });
          }
        }, this);
        first = level[0].toUpperCase();
        if (!moduleLogger[first]) {
          return moduleLogger[first] = moduleLogger[level];
        }
      }, this));
      return moduleLogger;
    };
    Logger.prototype.dup = function(options) {
      var logger;
      if (options == null) {
        options = {};
      }
      logger = new Logger(this._opts, this._attrs, this._outputs);
      return logger;
    };
    Logger.prototype.create = function(options) {
      return new Logger(options);
    };
    return Logger;
  })();
  module.exports = defaultLogger = new Logger;
}).call(this);
